#include <iostream>
#include <stack>
#include <iterator>

template<typename T>
class BinaryTree {
private:
    struct Node {
        T data;
        Node* left;
        Node* right;
        Node(const T& val) : data(val), left(nullptr), right(nullptr) {}
    };

    Node* root;

public:
    BinaryTree() : root(nullptr) {}

    void insert(const T& value) {
        root = insertRec(root, value);
    }

    class Iterator {
    private:
        std::stack<Node*> stack;

        void pushLeft(Node* node) {
            while (node) {
                stack.push(node);
                node = node->left;
            }
        }

    public:
        using iterator_category = std::forward_iterator_tag;
        using value_type = T;
        using difference_type = std::ptrdiff_t;
        using pointer = T*;
        using reference = T&;

        Iterator(Node* root) {
            pushLeft(root);
        }

        Iterator& operator++() {
            Node* node = stack.top();
            stack.pop();
            pushLeft(node->right);
            return *this;
        }

        Iterator operator++(int) {
            Iterator tmp = *this;
            ++(*this);
            return tmp;
        }

        bool operator==(const Iterator& other) const {
            return (stack.empty() && other.stack.empty()) ||
                   (!stack.empty() && !other.stack.empty() && stack.top() == other.stack.top());
        }

        bool operator!=(const Iterator& other) const {
            return !(*this == other);
        }

        T& operator*() {
            return stack.top()->data;
        }
    };

    Iterator begin() {
        return Iterator(root);
    }

    Iterator end() {
        return Iterator(nullptr);
    }

private:
    Node* insertRec(Node* node, const T& value) {
        if (node == nullptr) {
            return new Node(value);
        }
        if (value < node->data) {
            node->left = insertRec(node->left, value);
        } else if (value > node->data) {
            node->right = insertRec(node->right, value);
        }
        return node;
    }
};

int main() {
    BinaryTree<int> tree;
    tree.insert(5);
    tree.insert(3);
    tree.insert(7);
    tree.insert(1);
    tree.insert(9);

    std::cout << "In-order traversal: ";
    for (const auto& value : tree) {
        std::cout << value << " ";
    }
    std::cout << std::endl;

    // Using the iterator with STL algorithms
    auto minElement = std::min_element(tree.begin(), tree.end());
    auto maxElement = std::max_element(tree.begin(), tree.end());

    std::cout << "Min element: " << *minElement << std::endl;
    std::cout << "Max element: " << *maxElement << std::endl;

    return 0;
}
